import numpy as np
import os
import SimpleITK as sitk


def resampling(modified_path, reference_path, save_folder, interpolator=sitk.sitkLinear):
    os.makedirs(save_folder, exist_ok=True)
    info = os.path.basename(reference_path).split('_')
    filename = info[0] + '_' + info[1]
    modified_nii_img = sitk.ReadImage(modified_path, imageIO="NiftiImageIO")
    reference_nii_img = sitk.ReadImage(reference_path, imageIO="NiftiImageIO")
    resampler = sitk.ResampleImageFilter()
    resampler.SetOutputOrigin(modified_nii_img.GetOrigin())
    resampler.SetOutputDirection(modified_nii_img.GetDirection())
    resampler.SetSize(reference_nii_img.GetSize())
    resampler.SetOutputSpacing(reference_nii_img.GetSpacing())
    resampler.SetInterpolator(interpolator)
    resampled_modified_nii_img = resampler.Execute(modified_nii_img)

    sitk.WriteImage(resampled_modified_nii_img, f'{save_folder}/{filename}_resampled_to_mask.nii')
    
    return resampled_modified_nii_img 


def resample_fixedsize_fixedspacing (nifti_file_path: str, save_folder_path: str, interpolation: str, new_size: int, new_spacing: list) -> None:
    '''
    #img_sitk : sitk image (3D)
    #interppolation: sitk.sitkLinear
    #new_size : 64
    #new_spacing: [1,1,1]
    
    Example Code:
      resample_fixedsize_fixedspacing(nii_list[0], './result', sitk.sitkLinear, 64, [1,1,1])
    '''
    img_sitk = sitk.ReadImage(nifti_file_path)
    # img_array = sitk.GetArrayFromImage(img_sitk)
    try:
        file_name = nifti_file_path.split('\\')[-1].rstrip('.nii.gz')
    except:
        file_name = nifti_file_path.split('/')[-1].rstrip('.nii.gz')
    dimension = img_sitk.GetDimension()
    reference_size = [new_size,new_size,new_size]
    reference_physical_size = np.zeros(img_sitk.GetDimension())
    reference_physical_size[:] = [(sz-1)*spc if sz*spc>mx  else mx for sz,spc,mx in zip(img_sitk.GetSize(), img_sitk.GetSpacing(), reference_physical_size)]
    reference_direction = img_sitk.GetDirection()
    reference_origin = img_sitk.GetOrigin()
    reference_image = sitk.Image(reference_size, img_sitk.GetPixelIDValue())
    reference_image.SetOrigin(reference_origin)
    reference_image.SetSpacing(new_spacing)
    reference_image.SetDirection(reference_direction)
    reference_center = np.array(reference_image.TransformContinuousIndexToPhysicalPoint(np.array(reference_image.GetSize())/2.0))
    transform = sitk.AffineTransform(dimension)
    transform.SetMatrix(img_sitk.GetDirection())
    transform.SetTranslation(np.array(img_sitk.GetOrigin()) - reference_origin)
    centering_transform = sitk.TranslationTransform(dimension)
    img_center = np.array(img_sitk.TransformContinuousIndexToPhysicalPoint(np.array(img_sitk.GetSize())/2.0))
    centering_transform.SetOffset(np.array(transform.GetInverse().TransformPoint(img_center) - reference_center))
    centered_transform = sitk.CompositeTransform([centering_transform,transform])
    new_img = sitk.Resample(img_sitk, reference_image, centered_transform, interpolation, 0.0)
    os.makedirs(save_folder_path, exist_ok=True)
    sitk.WriteImage(new_img, f'{save_folder_path}/{file_name}_resampled.nii.gz')

    return None


def make_isotropic(image, interpolator = sitk.sitkLinear, spacing = None):
    '''
    Many file formats (e.g. jpg, png,...) expect the pixels to be isotropic, same
    spacing for all axes. Saving non-isotropic data in these formats will result in
    distorted images. This function makes an image isotropic via resampling, if needed.
    Args:
        image (SimpleITK.Image): Input image.
        interpolator: By default the function uses a linear interpolator. For
                      label images one should use the sitkNearestNeighbor interpolator
                      so as not to introduce non-existant labels.
        spacing (float): Desired spacing. If none given then use the smallest spacing from
                         the original image.
    Returns:
        SimpleITK.Image with isotropic spacing which occupies the same region in space as
        the input image.
    '''
    original_spacing = image.GetSpacing()
    # Image is already isotropic, just return a copy.
    if all(spc == original_spacing[0] for spc in original_spacing):
        return sitk.Image(image)
    # Make image isotropic via resampling.
    original_size = image.GetSize()
    if spacing is None:
        spacing = min(original_spacing)
    new_spacing = [spacing]*image.GetDimension()
    new_size = [int(round(osz*ospc/spacing)) for osz, ospc in zip(original_size, original_spacing)]
    return sitk.Resample(image, new_size, sitk.Transform(), interpolator,
                         image.GetOrigin(), new_spacing, image.GetDirection(), 0, # default pixel value
                         image.GetPixelID())


import os
import numpy as np
import nibabel as nib
import matplotlib.pyplot as plt
from io import StringIO


def find_all_files(top_folder_path: str, include_word: str) -> list:
    all_file_pathes = []
    for (root, directories, files) in os.walk(top_folder_path):
        for file in files:
            if include_word in file:
                detect_file_path = os.path.join(root, file)
                all_file_pathes.append(detect_file_path)
    return all_file_pathes
    

def save_print_instance(*message):
    io = StringIO()
    print(*message, file=io, end="")
    return io.getvalue()


def center_niishow(nifti_file_path: str, axis: str) -> np.array:
    nii = nib.load(nifti_file_path)
    nii_array = nii.get_fdata()
    try:
        x,y,z = nii_array.shape
        if axis=='x':
            plt.imshow(nii_array[int(np.round(x/2)),:,:])
        elif axis=='y':
            plt.imshow(nii_array[:,int(np.round(y/2)),:])
        elif axis=='z':
            plt.imshow(nii_array[:,:,int(np.round(z/2))])
        
    except Exception as e:
        print(e)
        pass
    # print(nii_array.shape)
    # print(f'sx=={sx}, sy=={sy}, sz=={sz}, volume_unit_of_nii=={volume_unit}')

    return nii_array


# support func for <get_matching_path_dict> method 
def get_mask_nii_path(nii_list: list, subject_numb: int, mask_word: str) -> str:
    '''Support method for `get_matching_path_dict`

    parameters:
    -----------
    nii_list: list
        globbed nii list(raw and mask)
    subject_numb: int 
        only for BIDS format file name must be has subject and modality
    mask_word: str 
        mask file include word(like 'msk' or 'mask' word in mask nifti file name)
    '''
    mask_nii_path = None
    for nii_path in nii_list:
        if subject_numb in nii_path and mask_word in nii_path:
            mask_nii_path = nii_path

    return mask_nii_path


def get_matching_path_dict(nii_list: list, maskword: str) -> dict: 
    '''
    Example code:
        matching_path_dict = get_matching_path_dict(nii_list, 'msk')
        matching_path_dict_keys = [key for key in matching_path_dict.keys()]

    parameters:
    -----------
    nii_list: str 
        1 list containing globs of raw and mask filepaths in BIDS format
    maskword: 
        mask file include word(like 'msk' or 'mask' word in mask nifti file name)
    '''
    matching_path_dict = dict()
    for nii_path in nii_list:
        if maskword not in nii_path:
            raw_nii = nii_path
            subject = nii_path.split('\\')[-1].split('_')[0]
            mask_nii_path = get_mask_nii_path(nii_list, subject, maskword)
            matching_path_dict[raw_nii] = mask_nii_path
        else:
            pass
    
    return matching_path_dict


def get_maskout_array(resampled_raw_nifti_path: str, resampled_mask_nifti_path: str, axial: str) -> np.array:
    raw_nii = nib.load(resampled_raw_nifti_path)
    raw_array = raw_nii.get_fdata()
    x,y,z = raw_array.shape[0], raw_array.shape[1], raw_array.shape[2]
    mask_nii = nib.load(resampled_mask_nifti_path)
    mask_array = mask_nii.get_fdata()

    if axial == 'x':
        mask_array = mask_array[:x,:,:]
    elif axial == 'y':
        mask_array = mask_array[:,:y,:]
    elif axial == 'z':
        mask_array = mask_array[:,:,:z]

    if raw_array.shape == mask_array.shape:
        mask_array = 1-mask_array
        maskout_array = raw_array * mask_array
    else:
        maskout_array = None
        print('Error [get_maskout_array]: raw_nii shape does not match to mask nii shape. please check')

    return maskout_array


import sys
import numpy as np
import nibabel as nib
from pathlib import Path
from skimage import morphology


def voxel_erosion(array3d:np.array) -> np.array:
    (x, y, z) = np.shape(array3d)
    erosion_mask = array3d.copy()
    print(x,y,z)
    for slice_number in range(z):
        image_slice = erosion_mask[:, :, slice_number]
        erosion_mask[:, :, slice_number] = morphology.binary_erosion(image_slice, morphology.diamond(1))
        
    return erosion_mask        


def voxel_dilation(array3d: np.array) -> np.array:
    (x, y, z) = np.shape(array3d)
    dilated_img = array3d.copy()
    for slice_number in range(z):
        image_slice = dilated_img[:, :, slice_number]
        dilated_img[:, :, slice_number] = morphology.binary_dilation(image_slice, morphology.diamond(1))
        
    return dilated_img


def get_boundary_diff_index(raw_nifti_file_path, mask_nifti_file_path):
    raw_nii = nib.load(raw_nifti_file_path)
    raw_nii_array = raw_nii.get_fdata()
    x,y,z = raw_nii_array.shape
    # raw_affine = raw_nii.affine
    mask_nii = nib.load(mask_nifti_file_path)
    mask_nii_array = mask_nii.get_fdata()
    mask_nii_array = mask_nii_array[:,:,:z]
    # mask_affine = mask_nii.affine
    
    eroded_mask = voxel_erosion(mask_nii_array)
    dilated_mask = voxel_dilation(mask_nii_array)
    out_boundary_mask = dilated_mask - mask_nii_array
    out_boundary_mask[out_boundary_mask<0] = 0
    inner_boundary_mask = mask_nii_array - eroded_mask
    out_boundary_val = raw_nii_array*out_boundary_mask
    inner_boundary_val = raw_nii_array*inner_boundary_mask
    boundary_diff_index = np.sum(out_boundary_val) - np.sum(inner_boundary_val)
        
    return boundary_diff_index


import numpy as np
import glob
import pandas as pd
import nibabel as nib
import monai # MONAI cropforeground vs EDA-NIf get_nonzero3d


def count_center_voxel_labels(self, nifti_path: str, x_val: int, y_val: int, z_val):
    img = nib.load(nifti_path)
    x,y,z = img.shape
    center_x, center_y, center_z = int(np.round(x/2)), int(np.round(y/2)), int(np.round(z/2))
    n_count = np.count_nonzero 
    img_array = img.get_fdata()

    center_region_mask_label_count = n_count(img_array [center_x-x_val:center_x+x_val, center_y-y_val:center_y+y_val, center_z-z_val:center_z+z_val])

    return center_region_mask_label_count


## get_nonzero3d
def if_minus_return_0(value: float) -> float:
    if np.sign(value) == -1:
        value = 0
    return value


def get_nonzero3d(nifti_file_path: str, save_path: str, margin: int) -> None:
    file_name = nifti_file_path.split('\\')[-1].rstrip('nii.gz')
    nii = nib.load(nifti_file_path)
    nii_array = nii.get_fdata()
    original_affine = nii.affine
    # print(nii.shape)
    non_zeros = np.nonzero(nii_array>3)

    x_min = non_zeros[0].min()
    x_max = non_zeros[0].max()
    y_min = non_zeros[1].min()
    y_max = non_zeros[1].max()
    z_min = non_zeros[2].min()
    z_max = non_zeros[2].max()
    
    clip_x_min = if_minus_return_0(x_min-margin)
    clip_x_max = if_minus_return_0(x_max+margin)
    clip_y_min = if_minus_return_0(y_min-margin)
    clip_y_max = if_minus_return_0(y_max-margin)
    clip_z_min = if_minus_return_0(z_min-margin)
    clip_z_max = if_minus_return_0(z_max-margin)
    
    nonzero3darray = nii_array[clip_x_min:clip_x_max, clip_y_min:clip_y_max, clip_z_min:clip_z_max]
    img = nib.Nifti1Image(nonzero3darray, original_affine) 
    img.to_filename(f'{save_path}/{file_name}.nii.gz')
    
    return nonzero3darray


## for get nonzero 3darray shape from cohort dataset
def get_nonzero3d_shape(nifti_file_path: str) -> int:
    nii = nib.load(nifti_file_path)
    nii_array = nii.get_fdata()
    non_zeros = np.nonzero(nii_array>2)

    x_min = non_zeros[0].min()
    x_max = non_zeros[0].max()
    y_min = non_zeros[1].min()
    y_max = non_zeros[1].max()
    z_min = non_zeros[2].min()
    z_max = non_zeros[2].max()
    return x_min, x_max, y_min, y_max, z_min, z_max


## for crop same size when all data(with mask) is co-registerd and preprocessed.
def get_hardcrop(nifti_file_path: str, save_path: str, x_min: int, x_max: int, y_min:int, y_max: int, z_min: int, z_max: int) -> None:
    file_name = nifti_file_path.split('\\')[-1].rstrip('nii.gz')
    nii = nib.load(nifti_file_path)
    nii_array = nii.get_fdata()
    nii_array = nii_array[x_min:x_max, y_min:y_max, z_min:z_max]
    original_affine = nii.affine
    img = nib.Nifti1Image(nii_array, original_affine) 
    img.to_filename(f'{save_path}/{file_name}.nii.gz')
    return None


## monai.cropforeground
def threshold_at_two(x):
    # threshold at 2
    return x > 2


def monai_cropforeground(nifit_file_path:str, save_path: str):
    file_name = nifit_file_path.split('\\')[-1].rstrip('.nii.gz')
    img = nib.load(nifit_file_path)
    img_array = img.get_fdata()
    original_affine = img.affine
    cropper = monai.transforms.CropForeground(select_fn=threshold_at_two, margin=2)
    cropped_img = cropper(img_array)
    img = nib.Nifti1Image(cropped_img, original_affine) 
    img.to_filename(f'{save_path}/{file_name}.nii.gz')


def get_nonzero_xyz_of_nii(nifti_file_path: str) -> (int, int, int, int, int, int):
    nii = nib.load(nifti_file_path)
    nii_array = nii.get_fdata()
    # print(nii.shape)
    non_zeros = np.nonzero(nii_array>3)
    x_min = non_zeros[0].min()
    x_max = non_zeros[0].max()
    y_min = non_zeros[1].min()
    y_max = non_zeros[1].max()
    z_min = non_zeros[2].min()
    z_max = non_zeros[2].max()
    
    return x_min, x_max, y_min, y_max, z_min, z_max




import numpy as np
import SimpleITK as sitk
from ipywidgets import interact, fixed
import matplotlib.pyplot as plt


class RegistrationMetric:
    def __init__(self):
        self.ant_neighborhood_radius = 1

    def calculate(self, nii_file_path1: str, nii_file_path2: str) -> dict:
        img1 = sitk.ReadImage(nii_file_path1)
        img2 = sitk.ReadImage(nii_file_path2)
        metric_dict = dict()
        registration_method = sitk.ImageRegistrationMethod()
        for metric in [
            "mutual_info",
            "ANTS_neighborhood_corr",
            "corr",
            "joint_histogram_mutualInfo",
            "mean_square",
        ]:
            if metric == "mutual_info":
                registration_method.SetMetricAsMattesMutualInformation()
            elif metric == "ANTS_neighborhood_corr":
                registration_method.SetMetricAsANTSNeighborhoodCorrelation(
                    self.ant_neighborhood_radius
                )
            elif metric == "corr":
                registration_method.SetMetricAsCorrelation()
            elif metric == "joint_histogram_mutualInfo":
                registration_method.SetMetricAsJointHistogramMutualInformation()
            elif metric == "mean_square":
                registration_method.SetMetricAsMeanSquares()
                # for more infomation check this page 
                # https://simpleitk.org/SPIE2019_COURSE/04_basic_registration.html

            metric_dict[metric] = np.round(registration_method.MetricEvaluate(img1, img2), 4)

        return metric_dict

    # If you can use AntsPy, you can use this method.
    # def display_images_with_alpha(image_z: int, alpha: np.float_, fixed: ants.core.ANTsImage, moving: ants.core.ANTsImage) -> plt:
    #     img = (1.0 - alpha) * fixed[:, :, image_z] + alpha * moving[:, :, image_z]
    #     plt.imshow(sitk.GetArrayViewFromImage(img), cmap=plt.cm.Greys_r)
    #     plt.axis("off")
    #     plt.show()


#https://discourse.itk.org/t/how-can-i-catch-the-registered-slices-in-moving-resampled-and-save-them-in-a-folder/4170
    def display_images_with_alpha(image_z, alpha, fixed, moving):
        img = (1.0 - alpha)*fixed[:,:,image_z] + alpha*moving[:,:,image_z] 
        plt.imshow(sitk.GetArrayViewFromImage(img),cmap=plt.cm.Greys_r);
        plt.axis('off')
        plt.show()

    def vis_registerd(self, nii_file_path1: str, nii_file_path2: str):
        fixed_image = sitk.ReadImage(nii_file_path1)
        moving_image = sitk.ReadImage(nii_file_path2)
        interact(
            self.display_images_with_alpha,
            fixed_image_z=(0, fixed_image.GetSize()[2] - 1),
            moving_image_z=(0, moving_image.GetSize()[2] - 1),
            fixed_npa=fixed(sitk.GetArrayViewFromImage(fixed_image)),
            moving_npa=fixed(sitk.GetArrayViewFromImage(moving_image)),
        )